#!/usr/bin/python
#transport_udp_session.py
#
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006-2012
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#


import os
import time
import struct
import cStringIO
import random


from twisted.internet import reactor
from twisted.internet.task import LoopingCall 


import dhnio
import settings
import misc
import identity
import identitycache
import tmpfile 
import nameurl

import automat
import automats

#------------------------------------------------------------------------------ 

CODES = {'p:':  'PING',
         'g:':  'GREETING',
         'd:':  'DATA',
         'r:':  'REPORT',
         'a:':  'ALIVE',}

COMMANDS = {'PING':         'p:',
            'GREETING':     'g:',
            'DATA':         'd:',
            'REPORT':       'r:',
            'ALIVE':        'a:',}

BLOCK_SIZE_LEVELS = {   #0: 508,
                        0: 1460,
                        1: 2048,
                        2: 3072,
                        3: 4096,
                        4: 5120,
                        5: 6144,
                        6: 7168,
                        7: 8124,
                        }
MAX_BLOCK_SIZE_LEVEL = len(BLOCK_SIZE_LEVELS) - 1

#BLOCK_SIZE = 4096
TIME_OUT = 20
BLOCK_RETRIES = 5
MAX_WINDOW_SIZE = 32
MIN_WINDOW_SIZE = 1

_ParentObject = None
_SessionsDict = {}
_IdToAddress = {}
_AddressToId = {}

_SendStatusCallback = None
_ReceiveStatusCallback = None
_SendControlFunc = None
_ReceiveControlFunc = None
_RegisterTransferFunc = None
_UnRegisterTransferFunc = None

_Timer1sec = None
_Timer10sec = None
_Timer30sec = None

_SendingDelay = 0.01
_SendingTask = None

_OutboxQueueDelay = 0.01
_OutboxQueueTask = None

#------------------------------------------------------------------------------ 

def init(parent):
    global _ParentObject
    dhnio.Dprint(4, 'transport_udp_session.init')
    _ParentObject = parent
    reactor.callLater(0, process_outbox_queue)
    reactor.callLater(1, process_sending)
    StartTimers()
   
def shutdown():
    dhnio.Dprint(4, 'transport_udp_session.shutdown')
    StopTimers()
    
   
def A(address, event=None, arg=None):
    global _SessionsDict
    s = open_session(address, arg)
    if event is not None:
        s.automat(event, arg)
    return s
      

def open_session(address, arg):
    global _SessionsDict
    if _SessionsDict.has_key(address):
        return _SessionsDict[address]
    dhnio.Dprint(6, 'transport_udp_session.open_session making new entry for [%s], arg=%s' % (str(address), str(arg)))
    idurl = None
    idname = 'unknown'
    if isinstance(arg, str) and arg.startswith('http://'):
        idurl = arg
        idname = nameurl.GetName(idurl)
    if idurl is None:
        idurl_remote = identitycache.GetIDURLByIPPort(address[0], address[1])
        local_address = identitycache.RemapContactAddress(address)
        idurl_local = identitycache.GetIDURLByIPPort(local_address[0], local_address[1])
        if idurl_local:
            idurl = idurl_local
            dhnio.Dprint(6, 'transport_udp_session.open_session found idurl when remap local IP : ' + idurl)
        else:
            if idurl_remote:
                idurl = idurl_remote 
                dhnio.Dprint(6, 'transport_udp_session.open_session found idurl from IP and PORT : ' + idurl)
        idname = nameurl.GetName(idurl)
    name = 'udp_%s_(%s:%s)' % (idname, address[0], address[1])
    s = TransportUDPSession(address, name, 'AT_STARTUP', 12)
    _SessionsDict[address] = s
    s.event('init', idurl)
    return s
    
#------------------------------------------------------------------------------ 
    
class OutboxFile():
    def __init__(self, peer, remote_idurl, file_id, filename):
        global _RegisterTransferFunc
        self.peer = peer
        self.remote_idurl = remote_idurl
        self.file_id = file_id
        self.filename = filename
        try:
            self.size = os.path.getsize(self.filename)
        except:
            self.size = -1
        self.block_id = -1
        self.num_blocks = 0
        self.blocks = {} 
        self.bytes_sent = 0
        self.failed_blocks = {}
        self.has_failed_blocks = False
        self.transfer_id = None
        if _RegisterTransferFunc is not None:
            self.transfer_id = _RegisterTransferFunc('send', self.peer.remote_address, self.get_bytes_sent, self.filename, self.size)
        # dhnio.Dprint(6, 'transport_udp_session.OutboxFile [%d] to %s' % (self.file_id, nameurl.GetName(self.remote_idurl)))

    def close(self):
        global _UnRegisterTransferFunc
        if _UnRegisterTransferFunc is not None and self.transfer_id is not None:
            _UnRegisterTransferFunc(self.transfer_id)
        # dhnio.Dprint(6, 'transport_udp_session.OutboxFile close [%d] to %s (%s)' % (self.file_id, nameurl.GetName(self.remote_idurl), str(self.peer)))

    def get_bytes_sent(self):
        return self.bytes_sent 

    def report_block(self, block_id):
        if not self.blocks.has_key(block_id):
            # dhnio.Dprint(2, 'transport_udp_session.report_block WARNING unknown block_id in REPORT packet received from %s: [%d]' % (self.remote_address, block_id))
            return
        self.bytes_sent += len(self.blocks[block_id])
        del self.blocks[block_id]

    def send(self):
        do_send = False
        while len(self.peer.sliding_window) < self.peer.sliding_window_size and self.block_id < self.num_blocks and self.block_id >= 0:
            if not self.send_block(self.block_id):
                break
            self.peer.sliding_window[(self.file_id, self.block_id)] = time.time()
            self.block_id += 1
        return do_send
    
    def send_block(self, block_id):
        global _SendControlFunc
        data = self.blocks[block_id]
        if _SendControlFunc is not None:
            more_bytes = _SendControlFunc(self.peer.last_bytes_sent, len(data))
            if more_bytes < len(data):
                return False
        self.peer.last_bytes_sent = len(data)
        datagram = COMMANDS['DATA']
        datagram += struct.pack('i', self.file_id)
        datagram += struct.pack('i', block_id)
        datagram += struct.pack('i', self.num_blocks)
        datagram += struct.pack('i', len(data))
        datagram += data
        return self.peer.sendDatagram(datagram)

    def mark_failed_block(self, block_id):
        if not self.failed_blocks.has_key(block_id):
            self.failed_blocks[block_id] = 0
        self.failed_blocks[block_id] += 1
        if self.failed_blocks[block_id] >= BLOCK_RETRIES:
            self.has_failed_blocks = True
            # dhnio.Dprint(8, 'transport_udp_session.markFailedBlock (%d,%d) is failed to send' % (self.current_file_id, block_id))
            return True
        return False
    
    def read_blocks(self):
        fin = open(self.filename, 'rb')
        block_id = 0
        while True:
            block_data = fin.read(BLOCK_SIZE_LEVELS[self.peer.block_size_level])
            if block_data == '':
                break
            self.blocks[block_id] = block_data
            block_id += 1  
        fin.close()
        self.num_blocks = block_id
        self.block_id = 0     

#------------------------------------------------------------------------------ 


class InboxFile():
    def __init__(self, peer, file_id, num_blocks):
        global _RegisterTransferFunc
        self.fd, self.filename = tmpfile.make("udp-in")
        self.peer = peer
        self.file_id = file_id
        self.num_blocks = num_blocks
        self.blocks = {}
        self.bytes_received = 0
        self.transfer_id = None
        if _RegisterTransferFunc is not None:
            self.transfer_id = _RegisterTransferFunc('receive', self.peer.remote_address, self.get_bytes_received, self.filename, -1)
        # dhnio.Dprint(6, 'transport_udp_session.InboxFile {%s} [%d] from %s' % (self.transfer_id, self.file_id, str(self.peer.remote_address)))

    def close(self):
        global _UnRegisterTransferFunc
        if _UnRegisterTransferFunc is not None and self.transfer_id is not None:
            _UnRegisterTransferFunc(self.transfer_id)
        # dhnio.Dprint(6, 'transport_udp_session.InboxFile.close {%s} [%d] from %s' % (self.transfer_id, self.file_id, str(self.peer.remote_address)))

    def get_bytes_received(self):
        return self.bytes_received

    def input_block(self, block_id, block_data):
        self.blocks[block_id] = block_data
        self.bytes_received += len(block_data) 
    
    def build(self):
        for block_id in range(len(self.blocks)):
            os.write(self.fd, self.blocks[block_id])
        os.close(self.fd)
        # dhnio.Dprint(10, 'transport_udp_server.InboxFile.build [%s] file_id=%d, blocks=%d' % (
        #     os.path.basename(filename), file_id, len(self.incommingFiles[file_id])))
    
#------------------------------------------------------------------------------ 

class TransportUDPSession(automat.Automat):
    fast = True
#    timers = {'timer-1min':     (60,    ['CONNECTED', 'SENDING']),
#              'timer-30sec':    (30,    ['GREETING', 'PING', 'DISCONNECTED']),
#              'timer-1sec':     (1,     ['PING', 'GREETING', 'CONNECTED']),
#              'timer-01sec':    (0.02,   ['SENDING',])}
    def __init__(self, address, name, state, debug_level):
        global _ParentObject
        self.remote_address = address
        self.remote_idurl = None
        self.parent = _ParentObject
        self.client = _ParentObject.client
        self.last_alive_packet_time = 0
        self.incommingFiles = {}
        self.outgoingFiles = {}
        self.maxOutgoingFiles = 4
        self.receivedFiles = {}
        self.outbox_queue = []
        self.tries = 0
        self.sliding_window = {}
        self.sliding_window_size = MIN_WINDOW_SIZE
        self.last_bytes_sent = 0
        self.failedBlocks = {}
        self.hasFailedBlocks = False
        self.block_size_level = 0
        self.remote_id_request_last_time = 0
        automat.Automat.__init__(self, name, state, debug_level)
        
    def A(self, event, arg):
        #---CONNECTED---
        if self.state is 'CONNECTED':
            if event == 'datagram-received' :
                self.doReceiveData(arg)
            elif event == 'timer-30sec' and self.isAlive(arg) :
                self.doAlive(arg)
            elif event == 'timer-30sec' and not self.isAlive(arg) :
                self.state = 'DISCONNECTED'
            elif event == 'shutdown' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
        #---AT_STARTUP---
        elif self.state is 'AT_STARTUP':
            if event == 'init' :
                self.state = 'REMOTE_ID'
                self.doSaveIDURL(arg)
                self.doRequestRemoteID(arg)
        #---REMOTE_ID---
        elif self.state is 'REMOTE_ID':
            if event == 'remote-id-received' and not self.isIPPortChanged(arg) :
                self.state = 'PING'
                self.doPing(arg)
            elif event == 'remote-id-received' and self.isIPPortChanged(arg) :
                self.state = 'CLOSED'
                self.doTransportUDPNotify(arg)
                self.doDestroyMe(arg)
            elif event == 'remote-id-failed' :
                self.state = 'DISCONNECTED'
            elif event == 'datagram-received' :
                self.doReceiveData(arg)
            elif event == 'shutdown' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
        #---PING---
        elif self.state is 'PING':
            if event == 'datagram-received' and self.isPingOrGreeting(arg) :
                self.state = 'GREETING'
                self.doGreeting(arg)
            elif event == 'datagram-received' and not self.isPingOrGreeting(arg) :
                self.doReceiveData(arg)
            elif event == 'timer-1sec' :
                self.doPing(arg)
            elif event == 'timer-30sec' :
                self.state = 'DISCONNECTED'
            elif event == 'shutdown' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
        #---GREETING---
        elif self.state is 'GREETING':
            if event == 'datagram-received' and self.isGreetingOrAlive(arg) :
                self.state = 'CONNECTED'
                self.doAlive(arg)
            elif event == 'datagram-received' and not self.isGreetingOrAlive(arg) :
                self.doReceiveData(arg)
            elif event == 'timer-1sec' :
                self.doGreeting(arg)
            elif event == 'timer-30sec' :
                self.state = 'DISCONNECTED'
            elif event == 'shutdown' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
        #---DISCONNECTED---
        elif self.state is 'DISCONNECTED':
            if event == 'timer-10sec' :
                self.state = 'REMOTE_ID'
                self.doRequestRemoteID(arg)
            elif event == 'datagram-received' :
                self.state = 'PING'
                self.doPing(arg)
            elif event == 'shutdown' :
                self.state = 'CLOSED'
                self.doDestroyMe(arg)
        #---CLOSED---
        elif self.state is 'CLOSED':
            pass

    def isPingOrGreeting(self, arg):
        return arg[0][:2] in [ COMMANDS['PING'], COMMANDS['GREETING'], ] 
            
    def isGreetingOrAlive(self, arg):
        return arg[0][:2] in [ COMMANDS['GREETING'], COMMANDS['ALIVE'], ]
    
    def isAlive(self, arg):
        return time.time() - self.last_alive_packet_time < 60.0 * 1.0 + 5.0
    
    def isIPPortChanged(self, arg):
        if self.remote_idurl is None:
            return False
        ident = identitycache.FromCache(self.remote_idurl)
        new_address = ident.getProtoContact('udp')
        if new_address is None:
            return True
        try:
            ip, port = new_address[6:].split(':')
            new_address = (ip, int(port))
        except:
            return True
        new_address = self.parent.remapAddress(new_address)
        return self.remote_address != new_address                
    
    def doSaveIDURL(self, arg):
        self.remote_idurl = arg
  
    def doRequestRemoteID(self, arg):
        if self.remote_idurl is None:
            self.automat('remote-id-received')
        else:
            # we do not need to request remote ID too often
            if time.time() - self.remote_id_request_last_time < 60*5:
                self.automat('remote-id-received')
            else:
                self.remote_id_request_last_time = time.time()
                identitycache.immediatelyCaching(self.remote_idurl).addCallbacks(
                    lambda src: self.automat('remote-id-received', src),
                    lambda x: self.automat('remote-id-failed', x))
    
    def doPing(self, arg):
        self.sendDatagram(COMMANDS['PING'])
    
    def doGreeting(self, arg):
        self.sendDatagram(COMMANDS['GREETING']+misc.getLocalID())
    
    def doAlive(self, arg):
        self.sendDatagram(COMMANDS['ALIVE'])
    
    def doReceiveData(self, arg):
        global _ReceiveControlFunc
        io = cStringIO.StringIO(arg[0])
        code = io.read(2)
        cmd = CODES.get(code, None)
        if cmd is None:
            # dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING incorrect data from %s: [%s]' % (self.remote_address, code))
            return
        if _ReceiveControlFunc is not None:
            seconds_pause = _ReceiveControlFunc(len(arg[0]))
            if seconds_pause > 0:
                self.client.transport.pauseProducing()
                reactor.callLater(seconds_pause, self.client.transport.resumeProducing)
        # dhnio.Dprint(8, '                                UDP:[%s] from %s' % (cmd, self.remote_address))
        self.last_alive_packet_time = time.time()
        #--- DATA
        if cmd == 'DATA':
            try:
                file_id = struct.unpack('i', io.read(4))[0]
                block_id = struct.unpack('i', io.read(4))[0]
                num_blocks = struct.unpack('i', io.read(4))[0]
                data_size = struct.unpack('i', io.read(4))[0]
            except:
                dhnio.DprintException()
                return
            if not self.incommingFiles.has_key(file_id):
                if self.receivedFiles.has_key(file_id):
                    self.sendDatagram(COMMANDS['REPORT'] + struct.pack('i', file_id) + struct.pack('i', block_id))
                    return
                # self.incommingFiles[file_id] = {}
                self.incommingFiles[file_id] = InboxFile(self, file_id, num_blocks)
            inp_data = io.read()
            if len(inp_data) != data_size:
                dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING incorrect datagram received from %s, not full data' % str(self.remote_address))
                return
            self.incommingFiles[file_id].input_block(block_id, inp_data) 
            self.sendDatagram(COMMANDS['REPORT'] + struct.pack('i', file_id) + struct.pack('i', block_id))
            # dhnio.Dprint(10, 'transport_udp_session.doReceiveData [%d] (%d/%d) from %s' % (file_id, block_id, num_blocks, self.remote_address))
            if len(self.incommingFiles[file_id].blocks) == num_blocks:
                # dhnio.Dprint(2, 'transport_udp_session.doReceiveData new file [%d] (%d/%d) from %s' % (file_id, block_id, num_blocks, self.remote_address))
                self.incommingFiles[file_id].build()
                file_received(self.incommingFiles[file_id].filename, 'finished', 'udp', self.remote_address)
                self.incommingFiles[file_id].close()   
                del self.incommingFiles[file_id]   
                # self.incommingFiles.pop(file_id, None)
                self.receivedFiles[file_id] = time.time()
                self.eraseOldFileIDs()
        #--- REPORT
        elif cmd == 'REPORT':
            try:
                file_id = struct.unpack('i', io.read(4))[0]
            except:
                dhnio.DprintException()
                return
            if not self.outgoingFiles.has_key(file_id):
                # dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING unknown file_id in REPORT packet received from %s: [%d]' % (self.remote_address, file_id))
                return
            try:
                block_id = struct.unpack('i', io.read(4))[0]
            except:
                dhnio.DprintException()
                return
            self.outgoingFiles[file_id].report_block(block_id)
            self.sliding_window.pop((file_id, block_id), None)
            self.sliding_window_size += 1
            if self.sliding_window_size > MAX_WINDOW_SIZE:
                self.sliding_window_size = MAX_WINDOW_SIZE
            if len(self.outgoingFiles[file_id].blocks) == 0:
                self.reportSentDone(file_id)
                self.closeOutboxFile(file_id)
            # dhnio.Dprint(8, 'transport_udp_session.doReceiveData REPORT on [%d,%d] received, blocks=%d, window=%d' % (file_id, block_id, len(self.outgoingFiles[file_id]), self.sliding_window_size))
        #--- GREETING
        elif cmd == 'GREETING':
            greeting_idurl = io.read()
            if greeting_idurl:
                # dhnio.Dprint(6, 'transport_udp_session.doReceiveData got remote idurl from GREETING: %s' % greeting_idurl)
                if self.remote_idurl is None:
                    self.remote_idurl = greeting_idurl
                    dhnio.Dprint(6, 'transport_udp_session.doReceiveData recognized remote idurl from GREETING: %s' % greeting_idurl)
                else:
                    if self.remote_idurl != greeting_idurl:
                        dhnio.Dprint(2, 'transport_udp_session.doReceiveData WARNING incorrect remote idurl in GREETING: %s' % greeting_idurl)
            self.sendDatagram(COMMANDS['ALIVE'])
        #--- PING
        elif cmd == 'PING':
            self.sendDatagram(COMMANDS['GREETING']+misc.getLocalID())
        #--- ALIVE            
        elif cmd == 'ALIVE':
            pass

    def doDestroyMe(self, arg):
        global _SessionsDict
        _SessionsDict.pop(self.remote_address)
        automats.get_automats_by_index().pop(self.index)
        
    def doTransportUDPNotify(self, arg):
        self.parent.A('ip-port-changed', self.remote_idurl)

    #------------------------------------------------------------------------------ 


    def sendDatagram(self, data):
        # dhnio.Dprint(10, '    [UDP] %d bytes to %s (%s)' % (len(data), self.remote_address, nameurl.GetName(self.remote_idurl)))
        if self.client is not None:
            try:
                self.client.transport.write(data, self.remote_address)
                return True
            except:
                return False 
        else:
            dhnio.Dprint(2, 'transport_udp_session.sendDatagram WARNING client is None')
            return False      
        
    def makeFileID(self):
        return int(str(int(time.time() * 100.0))[4:])   
    
    def addOutboxFile(self, filename):
        self.outbox_queue.append(filename)
        # dhnio.Dprint(6, 'transport_udp_session.addOutboxFile [%s], total outbox files: %d' % (os.path.basename(filename), len(self.outbox_queue)))
        
    def putOutboxFile(self, filename):
        self.outbox_queue.insert(0, filename) 
        # dhnio.Dprint(6, 'transport_udp_session.putOutboxFile [%s], total outbox files: %d' % (os.path.basename(filename), len(self.outbox_queue)))
    
    def eraseOldFileIDs(self):
        if len(self.receivedFiles) > 10:
            file_ids = self.receivedFiles.keys()
            cur_tm = time.time()
            for file_id in file_ids:
                if cur_tm - self.receivedFiles[file_id] > 60 * 5:
                    del self.receivedFiles[file_id]
            del file_ids 

    def closeOutboxFile(self, file_id):
        self.outgoingFiles[file_id].close()
        del self.outgoingFiles[file_id]

    def reportSentDone(self, file_id):
        file_sent(self.remote_address, self.outgoingFiles[file_id].filename, 'finished', 'udp')
        self.block_size_level += 1
        if self.block_size_level > MAX_BLOCK_SIZE_LEVEL:
            self.block_size_level = MAX_BLOCK_SIZE_LEVEL
    
    def reportSentFailed(self, file_id):
        file_sent(self.remote_address, self.outgoingFiles[file_id].filename, 'failed', 'udp')
        self.block_size_level -= 2
        if self.block_size_level < 0:
            self.block_size_level = 0

    def processOutboxQueue(self):
        has_reads = False
        while len(self.outbox_queue) > 0 and len(self.outgoingFiles) < self.maxOutgoingFiles:        
            filename = self.outbox_queue.pop(0)
            file_id = self.makeFileID()
            has_reads = True
            # we have a queue of files to be sent
            # somehow file may be removed before we start sending it
            # so we check it here and skip not existed files
            if not os.path.isfile(filename):
                file_sent(self.remote_address, filename, 'failed', 'udp', None, 'file were removed')
                continue
#            if os.access(filename, os.R_OK):
#                file_sent(self.remote_address, filename, 'failed', 'udp', None, 'access denied')
#                continue
            f = OutboxFile(self, self.remote_idurl, file_id, filename)
            f.read_blocks()
            self.outgoingFiles[file_id] = f
        return has_reads

    def processSending(self):
        has_sends = False
        failed_ids = []
        for file_id in self.outgoingFiles.keys():
            if self.outgoingFiles[file_id].send():
                has_sends = True
            if self.outgoingFiles[file_id].has_failed_blocks:
                failed_ids.append(file_id)
        for file_id in failed_ids:
            self.reportSentFailed(file_id)
            self.closeOutboxFile(file_id)
        del failed_ids
        self.checkWindow()
        return has_sends

    def checkWindow(self):
        if len(self.sliding_window) == 0:
            return
        to_remove = [] 
        to_update = {}
        cur_tm = time.time()
        for block_info, tm in self.sliding_window.items():
            if cur_tm - tm < TIME_OUT: # give some time to get an Ack. 
                continue
            to_remove.append(block_info)
            file_id = block_info[0]
            block_id = block_info[1]
            if not self.outgoingFiles.has_key(file_id):
                # dhnio.Dprint(2, 'transport_udp_session.checkWindow WARNING unknown file_id [%d] from %s' % (file_id, nameurl.GetName(self.remote_idurl)))
                continue
            if not self.outgoingFiles[file_id].blocks.has_key(block_id):
                dhnio.Dprint(2, 'transport_udp_session.checkWindow WARNING unknown block_id: %d' % block_id)
                continue
            self.outgoingFiles[file_id].send_block(block_id)
            isFailed = self.outgoingFiles[file_id].mark_failed_block(block_id)
            if not isFailed:
                to_update[block_info] = cur_tm
        for block_info in to_remove:
            del self.sliding_window[block_info]
            # dhnio.Dprint(8, 'transport_udp_session.doCheckWindow (%d,%d) removed, window=%d' % (file_id, block_id, len(self.sliding_window)))
        for block_info, tm in to_update.items():
            self.sliding_window[block_info] = tm
            # dhnio.Dprint(8, 'transport_udp_session.doCheckWindow (%s) resending to %s, window=%d' % (block_info, self.remote_address, len(self.sliding_window)))
        if len(to_remove) > 0:
            self.sliding_window_size = int( self.sliding_window_size / 2.0 )
            if self.sliding_window_size < MIN_WINDOW_SIZE:
                self.sliding_window_size = MIN_WINDOW_SIZE
            # dhnio.Dprint(8, 'transport_udp_session.doCheckWindow decreased, window=%d' % self.sliding_window_size)
        del to_remove
        del to_update

#------------------------------------------------------------------------------ 
    
def check_outbox_queue():    
    global _SessionsDict
    has_reads = False
    for sess in _SessionsDict.values():
        if sess.processOutboxQueue():
            has_reads = True
    return has_reads 
    
def process_outbox_queue():
    global _OutboxQueueDelay
    global _OutboxQueueTask
    
    has_reads = check_outbox_queue()

    _OutboxQueueDelay = misc.LoopAttenuation(
        _OutboxQueueDelay, 
        has_reads, 
        settings.MinimumSendingDelay(), 
        settings.MaximumSendingDelay(),)
    
    # attenuation
    _OutboxQueueTask = reactor.callLater(_OutboxQueueDelay, process_outbox_queue)
        

def process_sending():
    global _SendingDelay
    global _SendingTask
    global _SessionsDict
    
    has_sends = False
    for sess in _SessionsDict.values():
        if sess.processSending():
            has_sends = True
    
    _SendingDelay = misc.LoopAttenuation(
        _SendingDelay, 
        has_sends, 
        settings.MinimumSendingDelay(), 
        settings.MaximumSendingDelay(),)
    
    # attenuation
    _SendingTask = reactor.callLater(_SendingDelay, process_sending)
    
#------------------------------------------------------------------------------ 

def sessions():
    global _SessionsDict
    return _SessionsDict.keys()


def is_session_opened(address, idurl=None):
    global _SessionsDict
    if _SessionsDict.has_key(address):
        return True
    return False
        

def outbox_file(address, idurl, filename, fast=False):
    global _SessionsDict
#    dhnio.Dprint(6, 'transport_udp_session.outbox_file %s to %s (%s)' % (
#        os.path.basename(filename), nameurl.GetName(idurl), str(address)))
    s = open_session(address, idurl)
    if fast:
        s.putOutboxFile(filename)
    else:
        s.addOutboxFile(filename)
    check_outbox_queue()


def big_event(event, arg=None):
    global _SessionsDict
    for address, A in _SessionsDict.items():
        A.event(event, arg)
        # A(address, event, arg)


def data_received(datagram, address):
    try:
        from transport_control import black_IPs_dict
        if address[0] in black_IPs_dict().keys():
            # dhnio.Dprint(12, 'transport_udp_session.data_received %d bytes from BLACK IP: %s, %s' % (len(datagram), str(address[0]), str(black_IPs_dict()[address[0]])) )
            return
    except:
        pass
    if address not in sessions():
        open_session(address, None)
    A(address, 'datagram-received', (datagram, address))


def shutdown_all():
    big_event('shutdown')
    
    
def file_received(filename, status, proto='', host=None, error=None, message=''):
    global _ReceiveStatusCallback
    if _ReceiveStatusCallback is not None:
        _ReceiveStatusCallback(filename, status, proto, host, error, message)
    
    
def file_sent(host, filename, status, proto='', error=None, message=''):
    global _SendStatusCallback
    if _SendStatusCallback is not None:
        _SendStatusCallback(host, filename, status, proto, error, message)


def timer30sec():
    global _Timer30sec
    big_event('timer-30sec')
    _Timer30sec = reactor.callLater(random.randint(25,35), timer30sec)


def timer10sec():
    global _Timer10sec
    big_event('timer-10sec')
    _Timer10sec = reactor.callLater(random.randint(5,15), timer10sec)


def timer1sec():
    global _Timer1sec
    big_event('timer-1sec')
    _Timer1sec = reactor.callLater(1, timer1sec)
    
#------------------------------------------------------------------------------ 
    
def SetReceiveStatusCallback(cb):
    global _ReceiveStatusCallback
    _ReceiveStatusCallback = cb
    
    
def SetSendStatusCallback(cb):
    global _SendStatusCallback
    _SendStatusCallback = cb
    
def SetSendControlFunc(f):
    global _SendControlFunc
    _SendControlFunc = f
    
def SetReceiveControlFunc(f):
    global _ReceiveControlFunc
    _ReceiveControlFunc = f

def SetRegisterTransferFunc(f):
    global _RegisterTransferFunc
    _RegisterTransferFunc = f

def SetUnRegisterTransferFunc(f):
    global _UnRegisterTransferFunc
    _UnRegisterTransferFunc = f
    
#------------------------------------------------------------------------------ 

    
def StartTimers():
    global _Timer30sec
    global _Timer10sec
    global _Timer1sec
    if _Timer30sec is None:
        timer30sec()
    if _Timer10sec is None: 
        timer10sec()
    if _Timer1sec is None:
        timer1sec()


def StopTimers():
    global _Timer30sec
    global _Timer10sec
    global _Timer1sec
    if _Timer30sec:
        _Timer30sec.cancel()
        _Timer30sec = None
    if _Timer10sec:
        _Timer10sec.cancel()
        _Timer10sec = None
    if _Timer1sec:
        _Timer1sec.cancel()
        _Timer1sec = None








#!/usr/bin/python
#block_rebuilder.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006-2012
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#

import os
import sys
import time
import random


try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in block_rebuilder.py')

from twisted.internet.defer import Deferred, maybeDeferred
from twisted.internet import threads


import lib.dhnio as dhnio
import lib.misc as misc
import lib.nameurl as nameurl
import lib.transport_control as transport_control
import lib.settings as settings
import lib.contacts as contacts
import lib.eccmap as eccmap
import lib.tmpfile as tmpfile
import lib.diskspace as diskspace
import lib.packetid as packetid
from lib.automat import Automat


import fire_hire
import backup_rebuilder
import list_files_orator 
import lib.automats as automats

import backups
import raidread
import p2p_service
import io_throttle
import backup_db
import contact_status
import data_sender

_BlockRebuilder = None

#------------------------------------------------------------------------------ 

def A(event=None, arg=None):
    global _BlockRebuilder
    if _BlockRebuilder is None:
        _BlockRebuilder = BlockRebuilder()
    if event is not None:
        _BlockRebuilder.automat(event, arg)
    return _BlockRebuilder


class BlockRebuilder(Automat):
    timers = {'timer-1min':     (60,    ['REQUEST',]),
              'timer-1sec':     (1,     ['REQUEST', 'SENDING']),}
    
    def __init__(self,  
                 eccMap, 
                 backupID, 
                 blockNum, 
                 supplierSet, 
                 remoteData, 
                 remoteParity,
                 localData, 
                 localParity, 
                 creatorId = None, 
                 ownerId = None):
        self.eccMap = eccMap
        self.backupID = backupID
        self.blockNum = blockNum
        self.supplierSet = supplierSet
        self.supplierCount = len(self.supplierSet.suppliers)
        self.remoteData = remoteData
        self.remoteParity = remoteParity
        self.localData = localData
        self.localParity = localParity
        self.creatorId = creatorId
        self.ownerId = ownerId
        # at some point we may be dealing with when we're scrubbers
        if self.creatorId == None:
            self.creatorId = misc.getLocalID()
        if self.ownerId == None:
            self.ownerId = misc.getLocalID()
        # this files we want to rebuild
        # need to identify which files to work on
        self.missingData = [0] * self.supplierCount
        self.missingParity = [0] * self.supplierCount
        # list of packets ID we requested
        self.requestedFilesList = []
        # array to remember requested files
        self.requestedData = [0] * self.supplierCount
        self.requestedParity = [0] * self.supplierCount
        
        # list of files to be send
        # self.outstandingFilesList = []
        # time when we start sending files
        # self.sendingStartedTime = 0
        # timeout to send a single file to supplier,
        # *2 because we want to be sure that it is not handled by io_throttle - it counts timeout too
        # self.timeoutSending = 2 * int(settings.PacketSizeTarget() / settings.SendingSpeedLimit())
        # 1 if we sent a Data file to single supplier, 2 if get Ack from him, 3 if it was failed   
        # self.dataSent = [0] * self.supplierCount
        # same for Parity files
        # self.paritySent = [0] * self.supplierCount
        # for every packet we sending - remember if it was success or failed  
        # self.sentReports = {}
        
        Automat.__init__(self, 'block_rebuilder', 'AT_STARTUP', 10)
        
    def state_changed(self, oldstate, newstate):
        backup_rebuilder.A('block_rebuilder.state', newstate)
        backups.RepaintBackup(self.backupID)
    
    def A(self, event, arg):
        #---AT_STARTUP---
        if self.state is 'AT_STARTUP':
            if event == 'start' :
                self.state = 'MISSING'
                self.doIdentifyRemoteMissingFiles(arg)
        #---MISSING---
        elif self.state is 'MISSING':
            if event == 'missing-files-identified' and not self.isMissingFilesOnHand(arg) :
                self.state = 'REQUEST'
                self.doRequestLocalMissingFiles(arg)
            elif event == 'missing-files-identified' and self.isMissingFilesOnHand(arg) :
                self.state = 'REBUILDING'
                self.doRebuild(arg)
        #---REQUEST---
        elif self.state is 'REQUEST':
            if not self.isStopped(arg) and ( ( event == 'timer-1sec' and self.isAllFilesReceived(arg) ) or event == 'timer-1min' ) :
                self.state = 'REBUILDING'
                self.doRebuild(arg)
            elif event == 'timer-1sec' and self.isStopped(arg) :
                self.state = 'STOPPED'
                self.doDestroyMe(arg)
        #---REBUILDING---
        elif self.state is 'REBUILDING':
            if event == 'rebuilding-finished' :
                self.state = 'DONE'
                self.doWorkDoneReport(arg)
                data_sender.A('new-data')
                self.doDestroyMe(arg)
        #---STOPPED---
        elif self.state is 'STOPPED':
            pass
        #---DONE---
        elif self.state is 'DONE':
            pass

    def isStopped(self, arg):
        return backup_rebuilder.ReadStoppedFlag()
    
    def isMissingFilesOnHand(self, arg):
        for supplierNum in range(self.supplierCount):
            # if supplier do not have the Data but is on line 
            if self.missingData[supplierNum] == 1:
                # ... and we also do not have the Data 
                if self.localData[supplierNum] != 1:
                    # return False - will need request the file   
                    return False
            # same for Parity                
            if self.missingParity[supplierNum] == 1:
                if self.localParity[supplierNum] != 1:
                    return False
        #dhnio.Dprint(8, 'block_rebuilder.isMissingFilesOnHand return True')
        return True
        
    def isAllFilesReceived(self, arg):
        return len(self.requestedFilesList) == 0
    
#    def isAllFilesSent(self):
#        return ( 1 not in self.dataSent ) and ( 1 not in self.paritySent )
    
#    def isOutstandingFiles(self):
#        return len(self.outstandingFilesList) > 0
    
#    def isTimeoutSending(self):
#        result = time.time() - self.sendingStartedTime > self.timeoutSending
#        if result:
#            dhnio.Dprint(6, 'block_rebuilder.isTimeoutSending return True, timeout=%d' % self.timeoutSending)
#        return result
    
    def doIdentifyRemoteMissingFiles(self, arg):
        def do_identify():
            self.availableSuppliers = self.supplierSet.GetActiveArray()
            for supplierNum in range(self.supplierCount):
                if self.availableSuppliers[supplierNum] == 0:
                    continue
                # if remote Data file not exist and supplier is online
                # we mark it as missing and will try to rebuild this file and send to him
                if self.remoteData[supplierNum] != 1:
                    # mark file as missing  
                    self.missingData[supplierNum] = 1
                # same for Parity file
                if self.remoteParity[supplierNum] != 1:
                    self.missingParity[supplierNum] = 1
            return True
        maybeDeferred(do_identify).addCallback(
            lambda x: self.automat('missing-files-identified'))
        
    def doRequestLocalMissingFiles(self, arg):
        self.availableSuppliers = self.supplierSet.GetActiveArray()
        # at the moment I'm download
        # everything I have available and needed
        for supplierNum in range(self.supplierCount):
            # if supplier is not alive - we can't request from him           
            if self.availableSuppliers[supplierNum] == 0:
                continue
            supplierID = self.supplierSet.suppliers[supplierNum]      
            # if the remote Data exist and is available because supplier is on line,
            # but we do not have it on hand - do request  
            if self.remoteData[supplierNum] == 1 and self.localData[supplierNum] == 0:
                PacketID = self.BuildFileName(supplierNum, 'Data')
                io_throttle.QueueRequestFile(
                    self.FileReceived, 
                    self.creatorId, 
                    PacketID, 
                    self.ownerId, 
                    supplierID)
                self.requestedFilesList.append(PacketID)
                self.requestedData[supplierNum] = 1
            # same for Parity
            if self.remoteParity[supplierNum] == 1 and self.localParity[supplierNum] == 0:
                PacketID = self.BuildFileName(supplierNum, 'Parity')
                io_throttle.QueueRequestFile(
                    self.FileReceived, 
                    self.creatorId, 
                    PacketID, 
                    self.ownerId, 
                    supplierID)
                self.requestedFilesList.append(PacketID)
                self.requestedParity[supplierNum] = 1
        
    def doRebuild(self, arg):
#        threads.deferToThread(self.AttemptRepair).addCallback(
#            lambda x: self.automat('rebuilding-thread-finished'))
        maybeDeferred(self.AttemptRebuild).addCallback(
            lambda x: self.automat('rebuilding-finished'))
        
    def doDestroyMe(self, arg):
        automats.get_automats_by_index().pop(self.index)
        reactor.callLater(0, backup_rebuilder.RemoveBlockRebuilder, self)
    
#    def doRemoveBlockFiles(self):
#        # we want to remove files for this block 
#        # because we only need them during rebuilding
#        if settings.getGeneralLocalBackups() is True:
#            # if user set this in settings - he want to keep the local files
#            return
#        # ... user do not want to keep local backups
#        if settings.getGeneralWaitSuppliers() is True:
#            # but he want to be sure - all suppliers are green for long time
#            if contact_status.hasOfflineSuppliers() or time.time() - fire_hire.GetLastFireTime() < 24*60*60:
#                # some people are not there or we do not have stable team yet.
#                # do not remove the files because we need it to rebuild
#                return
#        dhnio.Dprint(2, 'block_rebuilder.doRemoveBlockFiles [%s] [%s]' % (self.remoteData, self.remoteParity))
#        for supplierNum in range(self.supplierCount):
#            # data_filename = os.path.join(tmpfile.subdir('data-par'), self.BuildFileName(supplierNum, 'Data'))
#            # parity_filename = os.path.join(tmpfile.subdir('data-par'), self.BuildFileName(supplierNum, 'Parity'))
#            data_filename = os.path.join(settings.getLocalBackupsDir(), self.BuildFileName(supplierNum, 'Data'))
#            parity_filename = os.path.join(settings.getLocalBackupsDir(), self.BuildFileName(supplierNum, 'Parity'))
#            if os.path.isfile(data_filename) and self.remoteData[supplierNum] == 1:
#                try:
#                    os.remove(data_filename)
#                    #dhnio.Dprint(6, '    ' + os.path.basename(data_filename))
#                except:
#                    dhnio.DprintException()
#            if os.path.isfile(parity_filename) and self.remoteParity[supplierNum] == 1:
#                try:
#                    os.remove(parity_filename)
#                    #dhnio.Dprint(6, '    ' + os.path.basename(parity_filename))
#                except:
#                    dhnio.DprintException()
        
    def doWorkDoneReport(self, arg):
        if backup_rebuilder.ReadStoppedFlag() is False:
            for supplierNum in range(self.supplierCount):
                # if local file exist and we know we did the request - this mean it is a new file 
                if self.localData[supplierNum] == 1 and self.requestedData[supplierNum] == 1:
                    backups.LocalFileReport(None, self.backupID, self.blockNum, supplierNum, 'Data')
                if self.localParity[supplierNum] == 1 and self.requestedParity[supplierNum] == 1:
                    backups.LocalFileReport(None, self.backupID, self.blockNum, supplierNum, 'Parity')

    #------------------------------------------------------------------------------ 

    def HaveAllData(self, parityMap):
        for segment in parityMap:
            if self.localData[segment] == 0:
                return False
        return True


    def AttemptRebuild(self):
        #dhnio.Dprint(6, 'block_rebuilder.AttemptRepair BEGIN')
        madeProgress = True
        while madeProgress:
            madeProgress = False
            # will check all data packets we have 
            for supplierNum in range(self.supplierCount):
                dataFileName = self.BuildRaidFileName(supplierNum, 'Data')
                # if we do not have this item on hands - we will reconstruct it from other items 
                if self.localData[supplierNum] == 0:
                    parityNum, parityMap = self.eccMap.GetDataFixPath(self.localData, self.localParity, supplierNum)
                    if parityNum != -1:
                        rebuildFileList = []
                        rebuildFileList.append(self.BuildRaidFileName(parityNum, 'Parity'))
                        for supplierParity in parityMap:
                            if supplierParity != supplierNum:
                                rebuildFileList.append(self.BuildRaidFileName(supplierParity, 'Data'))
                        raidread.RebuildOne(rebuildFileList, len(rebuildFileList), dataFileName)
                    if os.path.exists(dataFileName):
                        self.localData[supplierNum] = 1
                        madeProgress = True
                        dhnio.Dprint(8, 'block_rebuilder.AttemptRebuild made progress with supplier %d' % supplierNum)
                # now we check again if we have the data on hand after rebuild at it is missing - send it
                # but also check to not duplicate sending to this man   
                # now sending is separated, see the file data_sender.py          
                if self.localData[supplierNum] == 1 and self.missingData[supplierNum] == 1: # and self.dataSent[supplierNum] == 0:
                    dhnio.Dprint(8, "block_rebuilder.AttemptRebuild new Data for supplier %d" % supplierNum)
                    # self.outstandingFilesList.append((dataFileName, self.BuildFileName(supplierNum, 'Data'), supplierNum))
                    # self.dataSent[supplierNum] = 1
        # now with parities ...            
        for supplierNum in range(self.supplierCount):
            parityFileName = self.BuildRaidFileName(supplierNum, 'Parity')
            if self.localParity[supplierNum] == 0:
                parityMap = self.eccMap.ParityToData[supplierNum]
                if self.HaveAllData(parityMap):
                    rebuildFileList = []
                    for supplierParity in parityMap:
                        rebuildFileList.append(self.BuildRaidFileName(supplierParity, 'Data'))
                    raidread.RebuildOne(rebuildFileList, len(rebuildFileList), parityFileName)
                    if os.path.exists(parityFileName):
                        dhnio.Dprint(8, 'block_rebuilder.AttemptRebuild parity file found after rebuilding for supplier %d' % supplierNum)
                        self.localParity[supplierNum] = 1
            # so we have the parity on hand and it is missing - send it
            if self.localParity[supplierNum] == 1 and self.missingParity[supplierNum] == 1: # and self.paritySent[supplierNum] == 0:
                dhnio.Dprint(8, 'block_rebuilder.AttemptRebuild new Parity for supplier %d' % supplierNum)
                # self.outstandingFilesList.append((parityFileName, self.BuildFileName(supplierNum, 'Parity'), supplierNum))
                # self.paritySent[supplierNum] = 1
        return True


#    def SendFileAcked(self, packet, remoteID, packetID):
#        dhnio.Dprint(8, 'block_rebuilder.SendFileAcked to %s with %s' % (nameurl.GetName(remoteID), packetID))
#        try:
#            backupID, blockNum, supplierNum, dataORparity = packetID.split("-")
#            blockNum = int(blockNum)
#            supplierNum = int(supplierNum)
#            if dataORparity == 'Data':
#                self.missingData[supplierNum] = 0
#                self.remoteData[supplierNum] = 1
#                self.dataSent[supplierNum] = 2
#            elif dataORparity == 'Parity':
#                self.missingParity[supplierNum] = 0
#                self.remoteParity[supplierNum] = 1
#                self.paritySent[supplierNum] = 2
#            else:
#                dhnio.Dprint(8, "block_rebuilder.SendFileAcked WARNING odd PacketID? -" + str(packetID))
#            if self.sentReports.has_key(packetID):
#                dhnio.Dprint(2, 'block_rebuilder.SendFileAcked WARNING already got sending report for %s' % packetID)
#            self.sentReports[packetID] = 'acked'
#        except:
#            dhnio.DprintException()
#        self.automat('file-sent-report')


#    def SendFileFailed(self, creatorID, packetID, reason):
#        dhnio.Dprint(8, 'block_rebuilder.SendFileFailed to %s with %s (%s)' % (nameurl.GetName(creatorID), packetID, reason))
#        try:
#            backupID, blockNum, supplierNum, dataORparity = packetID.split("-")
#            blockNum = int(blockNum)
#            supplierNum = int(supplierNum)
#            if dataORparity == 'Data':
#                self.missingData[supplierNum] = 1
#                self.remoteData[supplierNum] = 0
#                self.dataSent[supplierNum] = 3
#            elif dataORparity == 'Parity':
#                self.missingParity[supplierNum] = 1
#                self.remoteParity[supplierNum] = 0
#                self.paritySent[supplierNum] = 3
#            else:
#                dhnio.Dprint(8, "block_rebuilder.SendFileFailed WARNING odd PacketID? -" + str(packetID))
#            if self.sentReports.has_key(packetID):
#                dhnio.Dprint(2, 'block_rebuilder.SendFileFailed WARNING already got sending report for %s' % packetID)
#            self.sentReports[packetID] = reason
#        except:
#            dhnio.DprintException()
#        self.automat('file-sent-report')


    def FileReceived(self, packet, state):
        if state in ['in queue', 'shutdown']:
            return
        
        if state == 'exist':
            packetID = packet
            filename = os.path.join(settings.getLocalBackupsDir(), packetID)
            try:
                self.requestedFilesList.remove(packetID)
            except:
                dhnio.DprintException()
                return
            try: 
                supplierNum = int((packetID.split("-"))[2])
                if packetID.endswith("-Data"):
                    self.localData[supplierNum] = 1
                elif packetID.endswith("-Parity"):
                    self.localParity[supplierNum] = 1
            except:
                dhnio.DprintException()
            return              

        if state != 'received':
            dhnio.Dprint(4, "block_rebuilder.FileReceived WARNING incorrect state [%s] for packet %s" % (str(state), str(packet)))
            return
            
        packetID = packet.PacketID
        try:
            self.requestedFilesList.remove(packetID)
        except:
            dhnio.DprintException()
            return
        filename = os.path.join(settings.getLocalBackupsDir(), packetID)
        if packet.Valid():
            dhnio.Dprint(8, "block_rebuilder.FileReceived %s, requestedFilesList length is %d" % (packetID, len(self.requestedFilesList)))
            # if we managed to rebuild a file 
            # before a request came in don't overwrite it
            if os.path.exists(filename):
                dhnio.Dprint(4, "block_rebuilder.FileReceived WARNING rewriting existed file" + filename)
                try: 
                    os.remove(filename)
                except:
                    dhnio.DprintException()
            dhnio.WriteFile(filename, packet.Payload)
            try: 
                supplierNum = int((packetID.split("-"))[2])
                if packetID.endswith("-Data"):
                    self.localData[supplierNum] = 1
                elif packetID.endswith("-Parity"):
                    self.localParity[supplierNum] = 1
            except:
                dhnio.DprintException()              
        else:
            # TODO 
            # if we didn't get a valid packet ... re-request it or delete it?
            dhnio.Dprint(8, "block_rebuilder.FileReceived WARNING " + filename + " is not a valid packet")
            try: 
                os.remove(filename)
            except:
                dhnio.DprintException()


    def BuildRaidFileName(self, supplierNumber, dataOrParity):
        # return os.path.join(tmpfile.subdir('data-par'), self.BuildFileName(supplierNumber, dataOrParity))
        return os.path.join(settings.getLocalBackupsDir(), self.BuildFileName(supplierNumber, dataOrParity))


    def BuildFileName(self, supplierNumber, dataOrParity):
        return packetid.MakePacketID(self.backupID, self.blockNum, supplierNumber, dataOrParity)
        # return self.backupID + "-" + str(self.blockNum) + "-" + str(supplierNumber) + "-" + dataOrParity

#------------------------------------------------------------------------------ 



#!/usr/bin/python
#backup.py
#
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006-2012
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#
#
#  This interfaces between a pipe from something like tar and the twisted code
#    for rest of DataHaven.NET
#  We see how many packets are waiting to be sent,
#    and if it is not too many, and we can make more, we make some more.
#
#  Main idea:
#     1) When a backup is started a backup object is created
#     2) We get a file descriptor for the process creating the tar archive
#     3) We always use select/poll before reading so we never block
#     4) We also poll to see if more needed.
#     5) We number/name blocks so can be sure what is what when we read back later
#     6) We call raidmake to split block and make parities
#     7) We put parts into dhnpackets and give these to transport_control


import os
import sys
import time
import cStringIO
import gc


try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in backup.py')

from twisted.internet import threads
from twisted.internet.defer import Deferred, maybeDeferred


import lib.dhnio as dhnio
import lib.misc as misc
import lib.dhnpacket as dhnpacket
import lib.contacts as contacts
import lib.commands as commands
import lib.settings as settings
import lib.packetid as packetid
import lib.nonblocking as nonblocking
import lib.eccmap as eccmap
import lib.dhncrypto as dhncrypto
import lib.tmpfile as tmpfile
from lib.automat import Automat
import lib.automats as automats


import data_sender
import fire_hire


import raidmake
import dhnblock
import io_throttle
import events
import backup_db
import backups


#-------------------------------------------------------------------------------

class backup(Automat):
    timers = {'timer-01sec':    (0.1, ['RUN', 'READ']),} 
    
    def __init__(self, backupID, pipe, blockSize=None, resultDeferred=None):
        self.backupID = backupID
        self.eccmap = eccmap.Current()
        self.pipe = pipe
        self.blockSize = blockSize
        if self.blockSize is None:
            self.blockSize = self.eccmap.nodes() * settings.PacketSizeTarget()
        self.ask4abort = False
        self.stateEOF = False
        self.stateReading = False
        #self.currentBlockData = ''
        self.currentBlockData = cStringIO.StringIO()
        self.currentBlockSize = 0
        self.blockNumber = 0
        self.dataSent = 0
        self.blocksSent = 0
        self.closed = False
        # self.sendingDelay = 0.1
        self.result = ''
        if resultDeferred is None:
            self.resultDefer = Deferred()
        else:
            self.resultDefer = resultDeferred
        self.packetResultCallback = None
        # need to give time for supplier to receive the packet
        # but do not wait too long because other suppliers receive same data faster
        # this is a number of seconds we give to supplier to receive our packet
        # self.PacketFailedTimeout = int(settings.PacketSizeTarget() / settings.SendingSpeedLimit())
        # (packetID, filename) queue for each supplier by his number
        # [suplierNumber](packetID, filename)
        # self.outstandingPackets = {}
        # (packetID, filename) that is currently sending for every supplier number
        # self.currentlySendingPackets = {} 
        # time when we start sending the packet for each supplier by his number    
        # self.sendingBeginTimers = {}
        # how many packets was failed to send for each supplier by his number
        # self.sentFailedCounts = {} 
        # failed and success packets for each block by its number
        # store both data and parity packets in this form
        # [blockNumber]['D','P'][success, failed]      
        # self.blocksResultDict = {}
        # timed out packets list for each supplier by his number
        # self.timedOutPackets = {}
        # (total bytes sent, total duration sending) to each supplier by his number
        # we count only time "during sending packet",
        # this starts from moment of start sending single packet 
        # and moment when we receive an ack on this packet 
        self.sendingStats = {}
        Automat.__init__(self, 'backup', 'AT_STARTUP', 10)
        self.automat('init')
        events.info('backup', '%s started' % self.backupID)
        dhnio.Dprint(6, 'backup.__init__ %s %s %d' % (self.backupID, self.eccmap, self.blockSize,))
        
    def A(self, event, arg):
        #---AT_STARTUP---
        if self.state is 'AT_STARTUP':
            if event == 'init' :
                self.state = 'RUN'
        #---RUN---
        elif self.state is 'RUN':
            if event == 'timer-01sec' and self.isAborted(arg) :
                self.state = 'ABORTED'
                self.doClose(arg)
                self.doReport(arg)
                self.doDestroyMe(arg)
            elif event == 'timer-01sec' and not self.isAborted(arg) :
                self.state = 'READ'
        #---READ---
        elif self.state is 'READ':
            if event == 'timer-01sec' and self.isPipeReady(arg) and not self.isEOF(arg) and not self.isReadingNow(arg) and not self.isBlockReady(arg) :
                self.doRead(arg)
            elif event == 'timer-01sec' and not self.isReadingNow(arg) and ( self.isBlockReady(arg) or self.isEOF(arg) ) :
                self.state = 'BLOCK'
                self.doBlock(arg)
        #---BLOCK---
        elif self.state is 'BLOCK':
            if event == 'block-ready' :
                self.state = 'RAID'
                self.doRaid(arg)
        #---RAID---
        elif self.state is 'RAID':
            if event == 'raid-done' and not self.isEOF(arg) :
                self.state = 'RUN'
                self.doBlockReport(arg)
                data_sender.A('new-data')
                self.doNewBlock(arg)
            elif event == 'raid-done' and self.isEOF(arg) :
                self.state = 'DONE'
                self.doBlockReport(arg)
                data_sender.A('new-data')
                self.doClose(arg)
                self.doReport(arg)
                self.doDestroyMe(arg)
        #---DONE---
        elif self.state is 'DONE':
            pass
        #---ABORTED---
        elif self.state is 'ABORTED':
            pass

    def isAborted(self, arg):
        return self.ask4abort
         
#    def isSentAll(self):
#        return len(self.blocksResultDict) == 0 and self.dataSent > 0
    
    def isPipeReady(self, arg):
        return self.pipe is not None and self.pipe.state() in [nonblocking.PIPE_CLOSED, nonblocking.PIPE_READY2READ]
    
    def isBlockReady(self, arg):
        return self.currentBlockSize >= self.blockSize
    
    def isEOF(self, arg):
        return self.stateEOF
    
    def isReadingNow(self, arg):
        return self.stateReading

    def doClose(self, arg):
        self.closed = True
        
    def doDestroyMe(self, arg):
        self.currentBlockData.close()
        del self.currentBlockData
        automats.get_automats_by_index().pop(self.index)
        reactor.callLater(0, backup_db.RemoveRunningBackupObject, self.backupID)
        collected = gc.collect()
        dhnio.Dprint(6, 'backup.doDestroyMe collected %d objects' % collected)

    def doReport(self, arg):
        if self.result == '':
            self.resultDefer.callback(self.backupID)
            events.info('backup', '%s done successfully' % self.backupID)
        elif self.result == 'abort':  
            self.resultDefer.callback(self.backupID+' abort')
            events.info('backup', '%s aborted' % self.backupID)
        else:
            self.resultDefer.errback(self.backupID)
            dhnio.Dprint(1, 'backup.doReport ERROR %s result is [%s]' % (self.backupID, self.result))
            events.info('backup', '%s failed: %s' % (self.backupID, self.result))

    def doRead(self, arg):
        def readChunk():
            size = self.blockSize - self.currentBlockSize
            if size < 0:
                dhnio.Dprint(1, "backup.readChunk ERROR eccmap.nodes=" + str(self.eccmap.nodes()))
                dhnio.Dprint(1, "backup.readChunk ERROR blockSize=" + str(self.blockSize))
                dhnio.Dprint(1, "backup.readChunk ERROR currentBlockSize=" + str(self.currentBlockSize))
                raise Exception('size < 0, blockSize=%s, currentBlockSize=%s' % (self.blockSize, self.currentBlockSize))
                return ''
            elif size == 0:
                return ''
            if self.pipe is None:
                raise Exception('backup.pipe is None')
                return ''
            if self.pipe.state() == nonblocking.PIPE_CLOSED:
                dhnio.Dprint(4, 'backup.readChunk the state is PIPE_CLOSED !!!!!!!!!!!!!!!!!!!!!!!!')
                return ''
            if self.pipe.state() == nonblocking.PIPE_READY2READ:
                newchunk = self.pipe.recv(size)
                if newchunk == '':
                    dhnio.Dprint(4, 'backup.readChunk pipe.recv() returned empty string')
                return newchunk
            dhnio.Dprint(1, "backup.readChunk ERROR pipe.state=" + str(self.pipe.state()))
            raise Exception('backup.pipe.state is ' + str(self.pipe.state()))
            return ''
        def readDone(data):
            #self.currentBlockData += data
            self.currentBlockData.write(data)
            self.currentBlockSize += len(data)
            self.stateReading = False
            if data == '':
                self.stateEOF = True
            #dhnio.Dprint(12, 'backup.readDone %d bytes' % len(data))
        self.stateReading = True
        maybeDeferred(readChunk).addCallback(readDone)

    def doBlock(self, arg):
        def _doBlock():
            # dhnio.Dprint(12, 'backup.doBlock blockNumber=%d size=%d atEOF=%s' % (self.blockNumber, self.currentBlockSize, self.stateEOF))
            #src = self.currentBlockData
            src = self.currentBlockData.getvalue()
            #self.currentBlockData.close()
            #del self.currentBlockData
            ret = dhnblock.dhnblock(
                misc.getLocalID(),
                self.backupID,
                self.blockNumber,
                dhncrypto.NewSessionKey(),
                dhncrypto.SessionKeyType(),
                self.stateEOF,
                src,)
            del src
            return ret
        maybeDeferred(_doBlock).addCallback(
            lambda block: self.automat('block-ready', block),)

    def doRaid(self, arg):
        newblock = arg
        dhnio.Dprint(8, 'backup.doRaid block=%d size=%d eof=%s ' % (
            self.blockNumber, self.currentBlockSize, str(self.stateEOF),))
        fileno, filename = tmpfile.make('backup')
        serializedblock = newblock.Serialize()
        blocklen = len(serializedblock)
        os.write(fileno, str(blocklen) + ":" + serializedblock)
        os.close(fileno)
        threads.deferToThread(raidmake.raidmake, 
                              filename, 
                              self.eccmap.name, 
                              self.backupID, 
                              self.blockNumber).addBoth(
                                  lambda outDir: self.automat('raid-done', newblock))
        del serializedblock

    def doBlockReport(self, arg):
        newblock = arg
        for supplierNum in range(self.eccmap.nodes()):
            for DataOrParity in ('Data', 'Parity'):
                backups.LocalFileReport(None, newblock.BackupID, newblock.BlockNumber, supplierNum, DataOrParity)

#    def doSend(self, arg):
#        newblock = arg
#        for supplierNum in range(self.eccmap.nodes()):
#            # we do all data then all parity
#            for DataOrParity in ('Data','Parity'):
#                if not self.outstandingPackets.has_key(supplierNum):
#                    self.outstandingPackets[supplierNum] = []
#                # packetId = newblock.BackupID + '-' + str(newblock.BlockNumber) + '-' + str(supplierNum) + '-' + DataOrParity
#                packetId = packetid.MakePacketID(newblock.BackupID, newblock.BlockNumber, supplierNum, DataOrParity)
#                # fullfilename = os.path.join(tmpfile.subdir('data-par'), packetId)
#                fullfilename = os.path.join(settings.getLocalBackupsDir(), packetId)
#                self.outstandingPackets[supplierNum].append((packetId, fullfilename))
#        # self.dataSent += len(self.currentBlockData)
#        self.dataSent += self.currentBlockSize
#        del newblock

    def doNewBlock(self, arg):
        self.dataSent += self.currentBlockSize
        self.blocksSent += 1
        #self.currentBlockData = ''
        self.currentBlockData.close()
        del self.currentBlockData
        self.currentBlockData = cStringIO.StringIO()
        self.currentBlockSize = 0
        self.blockNumber += 1
        if self.packetResultCallback is not None:
            self.packetResultCallback(self.backupID, None)
        
#    def doStartProcessOutgoingFiles(self):
#        reactor.callLater(0, self.processSending)     
        
    #------------------------------------------------------------------------------ 

    def abort(self):
        dhnio.Dprint(4, 'backup.abort id='+str(self.backupID))
        self.ask4abort = True
        self.result = 'abort'
        
#    def processSending(self):
#        if self.closed:
#            return
##        if len(self.outstandingPackets) > 0:
##            self.sendingDelay = 0.1
##        else:
##            if self.sendingDelay < 4.0:
##                self.sendingDelay *= 2.0
#        self.sendingDelay = misc.LoopAttenuation(
#             self.sendingDelay, 
#             len(self.outstandingPackets) > 0, 
#             settings.MinimumSendingDelay(), 
#             settings.MaximumSendingDelay())
#        # let's check all outstanding packets for all suppliers
#        for supplierNum in self.outstandingPackets.keys():
#            # if nothing to send for this supplier - skip
#            if len(self.outstandingPackets[supplierNum]) == 0:
#                continue
#            # if we already sending a file to this man
#            # need to check how long we are doing this
#            beginSendingTime = self.sendingBeginTimers.get(supplierNum, None)
#            if beginSendingTime is not None:
#                if time.time() - beginSendingTime < self.PacketFailedTimeout:
#                    # currently sending to the supplier - check next supplier
#                    continue
#                dhnio.Dprint(6, 'backup.processSending timeout sending for supplier %d' % supplierNum)
#                # so this packet is timed out
#                # and we increase failed counter for this man
#                if not self.sentFailedCounts.has_key(supplierNum):
#                    self.sentFailedCounts[supplierNum] = 0
#                self.sentFailedCounts[supplierNum] += 1
#                # remove this packet from currently sending packets
#                timedOutPacketID, filename = self.currentlySendingPackets.pop(supplierNum) 
#                # now we want to mark this packet as timed out
#                # so when packetAcked or packetFailed will be fired by io_throttle
#                # we will check and do not update block result 
#                if not self.timedOutPackets.has_key(supplierNum):
#                    self.timedOutPackets[supplierNum] = []
#                self.timedOutPackets[supplierNum].append(timedOutPacketID)
#                # remove this timer because we know now that packet is failed
#                self.sendingBeginTimers.pop(supplierNum)
#                # finally update block result 
#                self.blockResult(timedOutPacketID, False)
#            # now we know we did not sending to this supplier
#            # and there are some files to be sent to him, take one from his queue
#            packetId, fullfilename = self.outstandingPackets[supplierNum].pop(0)
#            # if we already have 2 failed packets sent to this man
#            # we do not want to send any more, because no trust to him
#            # backup_monitor will reconstruct this backup later
#            if self.sentFailedCounts.has_key(supplierNum):
#                #dhnio.Dprint(6, 'backup.processSending supplier %d failed counts is %d' % (supplierNum, self.sentFailedCounts[supplierNum]))
#                if self.sentFailedCounts[supplierNum] >= 2:
#                    self.blockResult(packetId, False)
#                    continue
#            # do sending  
#            io_throttle.QueueSendFile(
#                fullfilename,
#                packetId,
#                contacts.getSupplierID(supplierNum),
#                misc.getLocalID(),
#                self.packetAcked,
#                self.packetFailed,)
#            # save current packetId and filename
#            self.currentlySendingPackets[supplierNum] = (packetId, fullfilename) 
#            # remember current time for this supplier
#            self.sendingBeginTimers[supplierNum] = time.time()
#            #dhnio.Dprint(6, 'backup.processSending begin sending [%s] to supplier %d' % (packetId, supplierNum))
#        # call again later
#        # attenuation
#        reactor.callLater(self.sendingDelay, self.processSending)


#    def packetAcked(self, packet, ownerID, packetID):
#        if self.closed:
#            return
#
#        # take ID, block number, supplier number and type
#        backupID, blockNum, supplierNum, dataORparity = packet.PacketID.split('-')
#        blockNum = int(blockNum)
#        supplierNum = int(supplierNum)
#        DorP = dataORparity[0] # "D" or "P"
#        if backupID != self.backupID:
#            dhnio.Dprint(4, 'backup.packetAcked WARNING backup ID did not match')
#        
#        # if packet is timed out - skip
#        if self.timedOutPackets.has_key(supplierNum):
#            if packetID in self.timedOutPackets[supplierNum]:
#                # but remove it from this list to be smart
#                self.timedOutPackets[supplierNum].remove(packetID)
#                return
#        
#        # ... and remove from currently sending packets
#        currentPacketID, filename = self.currentlySendingPackets.pop(supplierNum)
#        if packetID != currentPacketID:
#            raise Exception('incorrect packetID=%s for supplier %d, currentlySendingPackets=%s' % (packetID, supplierNum, self.currentlySendingPackets))
#
#        # and also remove timer so we know that supplier is free and able to receive next packet
#        startTimePacket = self.sendingBeginTimers.pop(supplierNum)
#        # increase amount of total bytes sent to this man
#        # supplier puts number of received bytes in the Ack
#        # and we know how much he received in last packet of Data     
#        if not self.sendingStats.has_key(supplierNum):
#            self.sendingStats[supplierNum] = [0, 0]
#        self.sendingStats[supplierNum][0] += int(packet.Payload)
#        # remember the period of time how long we sent this packet
#        self.sendingStats[supplierNum][1] += time.time() - startTimePacket
#        # now we can remove the local file because remote supplier got it and ...
#        if settings.getGeneralLocalBackups() is False:
#            # ... user do not want to keep local backups
#            if settings.getGeneralWaitSuppliers() is True:
#                # but he want to be sure - all suppliers are green for long time
#                if time.time() - fire_hire.GetLastFireTime() > 24*60*60:
#                    # that is fine - we know those people for long. they are trusted.
#                    # now we do not need local file
#                    # fullfilename = os.path.join(tmpfile.subdir('data-par'), packetID)
#                    fullfilename = os.path.join(settings.getLocalBackupsDir(), packetID)
#                    if os.path.isfile(fullfilename):
#                        dhnio.Dprint(2, 'backup.packetAcked will remove [%s]' % fullfilename)
#                        try:
#                            os.remove(fullfilename)
#                        except:
#                            dhnio.DprintException()
#        
#        # update block result finally
#        self.blockResult(packetID, True)
#        # update GUI
#        if self.packetResultCallback is not None:
#            self.packetResultCallback(self.backupID, packet)
#        #dhnio.Dprsint(10, 'backup.packetAcked [%s] outstanding=%d' % (packetID, len(self.outstanding)))


#    def packetFailed(self, creatorID, packetID, why):
#        if self.closed:
#            return
#        
#        # take ID, block number, supplier number and type
#        backupID, blockNum, supplierNum, dataORparity = packetID.split('-')
#        blockNum = int(blockNum)
#        supplierNum = int(supplierNum)
#        DorP = dataORparity[0] # "D" or "P"
#        if backupID != self.backupID:
#            dhnio.Dprint(4, 'backup.packetFailed WARNING backup ID did not match')
#
#        #dhnio.Dprint(6, 'backup.packetFailed [%s] for supplier %d' % (packetID, supplierNum))
#        # if packet is timed out - skip
#        if self.timedOutPackets.has_key(supplierNum):
#            if packetID in self.timedOutPackets[supplierNum]:
#                # but remove it from this list to be smart
#                self.timedOutPackets[supplierNum].remove(packetID)
#                return
#            
#        # ... and remove from currently sending packets
#        currentPacketID, filename = self.currentlySendingPackets.pop(supplierNum)
#        if packetID != currentPacketID:
#            raise Exception('incorrect packetID=%s for supplier %d, currentlySendingPackets=%s' % (packetID, supplierNum, self.currentlySendingPackets))
#        # and also remove timer so we know that supplier is free and able to receive next packet
#        startTime = self.sendingBeginTimers.pop(supplierNum)
#        # increase failed counter
#        if not self.sentFailedCounts.has_key(supplierNum):
#            self.sentFailedCounts[supplierNum] = 0
#        self.sentFailedCounts[supplierNum] += 1
#
#        # update block result finally
#        self.blockResult(packetID, False)
#        # do notify gui
#        if self.packetResultCallback is not None:
#            self.packetResultCallback(self.backupID, None)


#    def blockResult(self, PacketID, Result):
#        backupID, blockNum, supplierNum, dataORparity = PacketID.split('-')
#        blockNum = int(blockNum)
#        supplierNum = int(supplierNum)
#        DorP = dataORparity[0] # "D" or "P"
#        if not self.blocksResultDict.has_key(blockNum):
#            self.blocksResultDict[blockNum] = {'D': [0, 0], 'P': [0, 0]}
#        self.blocksResultDict[blockNum][DorP][int(Result)] += 1
#        succesData = self.blocksResultDict[blockNum]['D'][1]
#        failedData = self.blocksResultDict[blockNum]['D'][0]
#        succesParity = self.blocksResultDict[blockNum]['P'][1]
#        failedParity = self.blocksResultDict[blockNum]['P'][0]
#        if succesData + failedData >= self.eccmap.nodes() and succesParity + failedParity >= self.eccmap.nodes():
#            self.blocksResultDict.pop(blockNum)
#            self.blocksSent += 1
#            if succesData < self.eccmap.DataNeeded():
#                dhnio.Dprint(4, 'backup.blockResult WARNING block %d-Data keeps only %d suppliers from %d!' % (
#                    blockNum, succesData, self.eccmap.nodes()))
#            if succesParity < self.eccmap.DataNeeded():
#                dhnio.Dprint(4, 'backup.blockResult WARNING block %d-Parity keeps only %d suppliers from %d!' % (
#                    blockNum, succesParity, self.eccmap.nodes()))
#            dhnio.Dprint(6, 'backup.blockResult %d data[%d/%d] parity[%d/%d] %d more blocks' % (
#                blockNum, succesData, self.eccmap.nodes(), succesParity, self.eccmap.nodes(), len(self.blocksResultDict)))
#        backups.RemoteFileReport(backupID, blockNum, supplierNum, dataORparity, Result)




#    def SetPacketResultCallback(self, cb):
#        self.packetResultCallback = cb
        
        
#    def GetStats(self):
#        return self.sendingStats 
        
    def empty(self):
        pass
        
    def empty1(self):
        pass
    




#!/usr/bin/python
#data_sender.py
#
#    Copyright DataHaven.NET LTD. of Anguilla, 2006-2012
#    Use of this software constitutes acceptance of the Terms of Use
#      http://datahaven.net/terms_of_use.html
#    All rights reserved.
#

import os
import sys

try:
    from twisted.internet import reactor
except:
    sys.exit('Error initializing twisted.internet.reactor in data_sender.py')

import lib.dhnio as dhnio
import lib.misc as misc
import lib.packetid as packetid
import lib.contacts as contacts
import lib.settings as settings
import lib.diskspace as diskspace
import lib.nameurl as nameurl
import lib.transport_control as transport_control
import lib.automat as automat
import lib.automats as automats

import io_throttle
import backups

_DataSender = None

#------------------------------------------------------------------------------ 

def A(event=None, arg=None):
    global _DataSender
    if _DataSender is None:
        _DataSender = DataSender('data_sender', 'READY', 6)
    if event is not None:
        _DataSender.automat(event, arg)
    return _DataSender


class DataSender(automat.Automat):
    timers = {'timer-5min':     (60*5,     ['READY']),
              'timer-1sec':     (2,        ['SENDING'])}

    def state_changed(self, oldstate, newstate):
        automats.set_global_state('DATASEND ' + newstate)

    def A(self, event, arg):
        #---AT_STARTUP---
        if self.state is 'AT_STARTUP':
            if event == 'init' :
                self.state = 'READY'
        #---READY---
        elif self.state is 'READY':
            if event == 'new-data' or event == 'timer-5min' or event == 'restart' :
                self.state = 'SCAN_BLOCKS'
                self.doScanAndQueue(arg)
        #---SCAN_BLOCKS---
        elif self.state is 'SCAN_BLOCKS':
            if event == 'scan-done' and self.isQueueEmpty(arg) :
                self.state = 'READY'
            elif event == 'scan-done' and not self.isQueueEmpty(arg) :
                self.state = 'SENDING'
        #---SENDING---
        elif self.state is 'SENDING':
            if event == 'restart' or ( ( event == 'block-acked' or event == 'block-failed' ) and self.isQueueEmpty(arg) ) :
                self.state = 'SCAN_BLOCKS'
                self.doScanAndQueue(arg)
            elif event == 'timer-1sec' :
                self.doPrintStats(arg)

    def isQueueEmpty(self, arg):
        return io_throttle.IsSendingQueueEmpty()
    
    def doScanAndQueue(self, arg):
        dhnio.Dprint(6, 'data_sender.doScanAndQueue')
        for backupID in misc.sorted_backup_ids(backups.local_files().keys()):
            packetsBySupplier = backups.ScanBlocksToSend(backupID)
            for supplierNum in packetsBySupplier.keys():
                supplier_idurl = contacts.getSupplierID(supplierNum)
                if not supplier_idurl:
                    dhnio.Dprint(2, 'data_sender.doScanAndQueue WARNING ?supplierNum? %s for %s' % (packetID, backupID))
                    continue
                if not io_throttle.OkToSend(supplier_idurl):
                    continue
                for packetID in packetsBySupplier[supplierNum]:
                    backupID_, blockNum, supplierNum_, dataORparity = packetid.BidBnSnDp(packetID)
                    if backupID_ != backupID:
                        dhnio.Dprint(2, 'data_sender.doScanAndQueue WARNING ?backupID? %s for %s' % (packetID, backupID))
                        continue
                    if supplierNum_ != supplierNum:
                        dhnio.Dprint(2, 'data_sender.doScanAndQueue WARNING ?supplierNum? %s for %s' % (packetID, backupID))
                        continue
                    if io_throttle.HasPacketInSendQueue(supplier_idurl, packetID):
                        continue
                    filename = os.path.join(settings.getLocalBackupsDir(), packetID)
                    io_throttle.QueueSendFile(
                        filename, 
                        packetID, 
                        supplier_idurl, 
                        misc.getLocalID(), 
                        self.packetAcked, 
                        self.packetFailed)
                    # dhnio.Dprint(6, '  %s for %s' % (packetID, backupID))
        self.automat('scan-done')
        
    def doPrintStats(self, arg):
        if dhnio.Debug(9):
            transfers = transport_control.current_transfers()
            bytes_stats = transport_control.current_bytes_transferred()
            s = ''
            for info in transfers:
                s += '%s ' % (diskspace.MakeStringFromBytes(bytes_stats[info.transfer_id]).replace(' ', '').replace('bytes', 'b'))
            dhnio.Dprint(0, 'transfers: ' + s[:120])

    def packetAcked(self, packet, ownerID, packetID):
        backupID, blockNum, supplierNum, dataORparity = packetid.BidBnSnDp(packetID)
        backups.RemoteFileReport(backupID, blockNum, supplierNum, dataORparity, True)
        self.automat('block-acked', packetID)
    
    def packetFailed(self, remoteID, packetID, why):
        backupID, blockNum, supplierNum, dataORparity = packetid.BidBnSnDp(packetID)
        backups.RemoteFileReport(backupID, blockNum, supplierNum, dataORparity, False)
        self.automat('block-failed', packetID)
    
    
        
        
        
        
        
        
        
        

